!+
! Module: mathutils_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module mathutils_mod

  use dal
  use param  
  use errorHandling
  use constants
  use sort_mod

  implicit none 

  interface makeHistogram
    module procedure makeHistogramEdgesReal, makeHistogramBoundsReal
    module procedure makeHistogramEdgesDble, makeHistogramBoundsDble
  end interface

  interface rebin
    module procedure rebinBasicReal, rebinWithWidthsReal, rebinWithEdgesReal
    module procedure rebinBasicDble, rebinWithWidthsDble, rebinWithEdgesDble
  end interface

  interface normalise
    module procedure normaliseReal
    module procedure normaliseDble
  end interface

  interface vectorCross
    module procedure vectorCrossReal
    module procedure vectorCrossDble
  end interface

  interface integrate
    module procedure integrate1dReal, integrateNdReal
!    module procedure integrate1dDble, integrateNdDble
  end interface

logical, parameter :: testFlag = .false.
  integer, parameter :: defaultNumIntegralSamples = 100
  type :: boundType
    integer :: type
    real(kind(1d0)) :: value
  end type boundType
  type :: intervalType
    type(boundType) :: lower, upper
  end type intervalType
  type :: intervalsType
    type(boundType) :: lower, upper
    integer :: index
  end type intervalsType
  type :: sequenceType
    type(boundType) :: bound
    logical         :: isLower
    integer         :: index
  end type sequenceType
!  integer, parameter :: OPEN = 1	! From dssutil. Check that the
!  integer, parameter :: CLOSED = 2	! match-up is as I have guess it.
  integer, parameter :: INCLUSIVE = 1
  integer, parameter :: EXCLUSIVE = 2
  integer, parameter :: UNDEFINED = 3

  private 

  public makeHistogram
  public rebin
  public interpolate
  public calcEllipse
  public calcAitoff
  public angstroms2eV
  public eV2angstroms
  public bbFluxPerHertz
  public bbFluxPerMetre

  public exponentialRandom
  public getExponentialRandoms
  public gaussianRandom
  public getGaussianRandoms
  public randomNumber
  public getRandomNumbers

  public normalise
  public vectorCross
  public integrate
  public newtonsMethod

contains 

  !---------------------------------------------------------------------
  subroutine makeHistogramBoundsReal(inVector, binContents, lEdge, rEdge&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)            :: inVector(:)
    integer, intent(out)           :: binContents(:)
    real,    intent(in)            :: lEdge, rEdge
    integer, intent(out), optional :: underflows, overflows, status

    integer :: hsize, i, localUnderflows, localOverflows, localStatus
    real :: frac
    real :: binEdges(size(binContents) + 1)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)

    do i = 1, hsize + 1
      frac = (i - 1) / real(hsize)
      binEdges(i) = lEdge * (1.0 - frac) + rEdge * frac
    end do

    call makeHistogramKernel(dble(inVector), binContents, dble(binEdges)&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramBoundsReal

  !---------------------------------------------------------------------
  subroutine makeHistogramEdgesReal(inVector, binContents, binEdges&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)            :: inVector(:)
    integer, intent(out)           :: binContents(:)
    real,    intent(in),  optional :: binEdges(:)
    integer, intent(out), optional :: underflows, overflows, status

    integer :: hsize, binEdge, localUnderflows, localOverflows, localStatus
    real :: frac, minx, maxx
    real, allocatable :: myBinEdges(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)
    minx = minval(inVector)
    maxx = maxval(inVector)

    allocate(myBinEdges(hsize + 1))
    if (present(binEdges)) then
      myBinEdges = binEdges(1:hsize + 1)

    else
      do binEdge = 1, hsize + 1
        frac = (binEdge - 1) / real(hsize)
        myBinEdges(binEdge) = minx * (1.0 - frac) + maxx * frac
      end do
    end if

    call makeHistogramKernel(dble(inVector), binContents, dble(myBinEdges)&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramEdgesReal

  !---------------------------------------------------------------------
  subroutine makeHistogramBoundsDble(inVector, binContents, lEdge, rEdge&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)),  intent(in)            :: inVector(:)
    integer, intent(out)           :: binContents(:)
    real(kind(1d0)),  intent(in)            :: lEdge, rEdge
    integer, intent(out), optional :: underflows, overflows, status

    integer :: hsize, i, localUnderflows, localOverflows, localStatus
    real(kind(1d0)) :: frac
    real(kind(1d0)) :: binEdges(size(binContents) + 1)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)

    do i = 1, hsize + 1
      frac = (i - 1) / real(hsize)
      binEdges(i) = lEdge * (1.0 - frac) + rEdge * frac
    end do

    call makeHistogramKernel(inVector, binContents, binEdges&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramBoundsDble

  !---------------------------------------------------------------------
  subroutine makeHistogramEdgesDble(inVector, binContents, binEdges&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)),  intent(in)            :: inVector(:)
    integer, intent(out)           :: binContents(:)
    real(kind(1d0)),  intent(in),  optional :: binEdges(:)
    integer, intent(out), optional :: underflows, overflows, status

    integer :: hsize, binEdge, localUnderflows, localOverflows, localStatus
    real(kind(1d0)) :: frac, minx, maxx
    real(kind(1d0)), allocatable :: myBinEdges(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)
    minx = minval(inVector)
    maxx = maxval(inVector)

    allocate(myBinEdges(hsize + 1))
    if (present(binEdges)) then
      myBinEdges = binEdges(1:hsize + 1)

    else
      do binEdge = 1, hsize + 1
        frac = (binEdge - 1) / real(hsize)
        myBinEdges(binEdge) = minx * (1.0 - frac) + maxx * frac
      end do
    end if

    call makeHistogramKernel(inVector, binContents, myBinEdges&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramEdgesDble

  !---------------------------------------------------------------------
  subroutine makeHistogramKernel(inVector, binContents, binEdges&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)),  intent(in)  :: inVector(:)
    integer, intent(out) :: binContents(:)
    real(kind(1d0)),  intent(in)  :: binEdges(:)
    integer, intent(out) :: underflows, overflows, status

    integer :: vsize, hsize, i, bin
    real(kind(1d0)) :: minx, maxx, x

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    vsize = size(inVector)
    hsize = size(binContents)

    if (vsize .le. 1) then
      status = -1 ! only 1 value.
  return
    end if

    if (hsize .le. 1) then
      status = -2 ! only 1 bin.
  return
    end if

    if (size(binEdges) .lt. hsize + 1) then
      status = -3
  return
    else if (size(binEdges) .gt. hsize + 1) then
      status = status + 1
    end if

    do i = 1, hsize
      if (binEdges(i) .ge. binEdges(i + 1)) then
        status = -4
  return
      end if
    end do

    minx = binEdges(1)
    maxx = binEdges(hsize + 1)

    binContents = 0
    underflows = 0
    overflows = 0
    do i = 1, vsize
      x = inVector(i)
      if (x .lt. minx) then
        if (underflows .ge. huge(underflows)) then
          status = -5
  return
        end if
        underflows = underflows + 1

      else if (x .gt. maxx) then
        if (overflows .ge. huge(overflows)) then
          status = -6
  return
        end if
        overflows = overflows + 1

      else
        do bin = hsize, 1, -1
          if (x .ge. binEdges(bin)) then
            if (binContents(bin) .ge. huge(binContents)) then
              status = -7
  return
            end if
            binContents(bin) = binContents(bin) + 1
        exit
          end if
        end do
      end if
    end do

  end subroutine makeHistogramKernel

  !---------------------------------------------------------------------
  subroutine calcEllipse(xOffset, yOffset, xSemi, ySemi, angleDeg&
  , xVals, yVals)
  !---------------------------------------------------------------------
!*** from swcat.
! follow by call pgpoly(numPoints, xvals, yvals) to draw the ellipse.

    implicit none 

    real,  intent(in) :: xOffset,&
                         yOffset,&
                         xSemi,&
                         ySemi,&
                         angleDeg
    real, intent(out) :: xVals(:),&
                         yVals(:)

    integer :: point, numPoints
    real :: pointAngle, cosWholeAngle, sinWholeAngle, x, y

    !-------------------------------------------------------------------
    ! Start:

!    cosWholeAngle = cos(angleDeg * myPiOn180)
!    sinWholeAngle = sin(angleDeg * myPiOn180)
    cosWholeAngle = cos(angleDeg * PiOn180)
    sinWholeAngle = sin(angleDeg * PiOn180)

    numPoints = min(size(xVals), size(yVals))

    xVals = 0.0
    yVals = 0.0

    do point = 1, numPoints
!      pointAngle = (point / real(numPoints)) * myTwoPi
      pointAngle = (point / real(numPoints)) * TwoPi
      x = xSemi * cos(pointAngle)
      y = ySemi * sin(pointAngle)
      xVals(point) = xOffset + x * cosWholeAngle + y * sinWholeAngle
      yVals(point) = yOffset + y * cosWholeAngle - x * sinWholeAngle
    end do

  end subroutine calcEllipse

  !---------------------------------------------------------------------
  subroutine calcAitoff(raDeg, decDeg, x, y)
  !---------------------------------------------------------------------
!*** could make the aitoff calc quicker by precalculating sin and cos dec etc.
!*** from swcat.
!*** include also drawAitoffGrid?

    implicit none 

    real, intent(in) :: raDeg, decDeg
    real, intent(out) :: x, y
    real :: ra, dec, r1, denom, cosDec

    !-------------------------------------------------------------------
    ! Start:

!    ra  = raDeg  * myPiOn180
!    dec = decDeg * myPiOn180
    ra  = raDeg  * PiOn180
    dec = decDeg * PiOn180

!    r1 = (ra - myPi) / 2.0
    r1 = (ra - Pi) / 2.0
    cosDec = cos(dec)
    denom = sqrt(1.0 + cosDec * cos(r1))
    x = 2.0 * cosDec * sin(r1) / denom
    y = sin(dec) / denom

  end subroutine calcAitoff

  !---------------------------------------------------------------------
  function exponentialRandom()
  !---------------------------------------------------------------------

    implicit none

    real :: exponentialRandom
    integer, parameter :: size = 1000
    real :: harvest(size)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getExponentialRandoms(harvest)
    end if
    exponentialRandom = harvest(i)
    i = i + 1
    if (i .gt. size) i = 1

  end function exponentialRandom

  !---------------------------------------------------------------------
  subroutine getExponentialRandoms(numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have an exponential probability
  ! distribution with a mean value of 1.0.

    implicit none

    real, intent(out) :: numbers(:)
    real, allocatable :: harvest(:)
    integer :: n
    integer :: number

    !-------------------------------------------------------------------
    ! Start:

    number = size(numbers)
    allocate(harvest(number))

    call getRandomNumbers(harvest)
    do n = 1, number
      numbers(n) = -log(1.0 - harvest(n)) ! the subtraction of the
      ! random number from 1.0 is to prevent the occurrence of zero-
      ! valued argument to log().
    end do

  end subroutine getExponentialRandoms

  !---------------------------------------------------------------------
  function gaussianRandom()
  !---------------------------------------------------------------------

    implicit none

    real :: gaussianRandom
    integer, parameter :: size = 1000
    real :: harvest(size)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getGaussianRandoms(harvest)
    end if
    gaussianRandom = harvest(i)
    i = i + 1
    if (i .gt. size) i = 1

  end function gaussianRandom

  !---------------------------------------------------------------------
  subroutine getGaussianRandoms(numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have an gaussian probability
  ! distribution.

    implicit none

    real, intent(out) :: numbers(:)
!    real, allocatable :: harvest(:)
    integer :: n, number
    real fac, rsq, v1, v2

    !-------------------------------------------------------------------
    ! Start:

    number = size(numbers)
!    allocate(harvest(number))

!    call getRandomNumbers(harvest)

    if (number / 2.0 .eq. floor(number / 2.0)) then ! even
      do n = 1, number - 1, 2
        do
          v1 = 2.0 * randomNumber() - 1.0
          v2 = 2.0 * randomNumber() - 1.0
          rsq = v1 * v1 + v2 * v2

        if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
        end do

        fac = sqrt(-2.0 * log(rsq) / rsq)
        numbers(n)     = v1 * fac
!print *, fac, v1, n, numbers(n)
        numbers(n + 1) = v2 * fac
!print *, fac, v2, n + 1, numbers(n + 1)
      end do

    else ! odd
      do n = 1, number - 2, 2
        do
          v1 = 2.0 * randomNumber() - 1.0
          v2 = 2.0 * randomNumber() - 1.0
          rsq = v1 * v1 + v2 * v2

        if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
        end do

        fac = sqrt(-2.0 * log(rsq) / rsq)
        numbers(n)     = v1 * fac
        numbers(n + 1) = v2 * fac
      end do

      do
        v1 = 2.0 * randomNumber() - 1.0
        v2 = 2.0 * randomNumber() - 1.0
        rsq = v1 * v1 + v2 * v2

      if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
      end do

      fac = sqrt(-2.0 * log(rsq) / rsq)
      numbers(number) = v1 * fac
    end if

  end subroutine getGaussianRandoms

  !---------------------------------------------------------------------
  function randomNumber()
  !---------------------------------------------------------------------

    implicit none

    real :: randomNumber
    integer, parameter :: size = 1000
    real :: harvest(size)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getRandomNumbers(harvest)
    end if
    randomNumber = harvest(i)
    i = i + 1
    if (i .gt. size) i = 1

  end function randomNumber

  !---------------------------------------------------------------------
  subroutine getRandomNumbers(numbers)
  !---------------------------------------------------------------------

    implicit none

    real, intent(out) :: numbers(:)

    !-------------------------------------------------------------------
    ! Start:

    call random_number(numbers)

  end subroutine getRandomNumbers

  !---------------------------------------------------------------------
  real function myran1(rseed)
  !---------------------------------------------------------------------
!**** from evarimgen

    implicit none

    integer rseed, ia, im, iq, ir, ntab, ndiv
    real am, eps, rnmx
    parameter (ia = 16807,&
     		   im = 2147483647,&
     		   am = 1./im,&
     		   iq = 127773,&
     		   ir = 2836,&
     		 ntab = 32,&
     		 ndiv = 1 + (im - 1) / ntab,&
     		  eps = 1.2e-7,&
     		 rnmx = 1. - eps)
    integer j, k, iv(ntab), iy
    save iv, iy
    data iv /ntab * 0/, iy /0/

    if (rseed.le.0.or.iy.eq.0) then ! initialise:
      rseed = max(-rseed, 1) ! Prevents rseed = 0.
      do j = ntab + 8, 1, -1
        k = rseed / iq
        rseed = ia * (rseed - k * iq) - ir * k
        if (rseed.lt.0) rseed = rseed + im
        if (j.le.ntab) iv(j) = rseed
      end do
      iy = iv(1)
    end if

    k = rseed / iq
    rseed = ia * (rseed - k * iq) - ir * k
    if (rseed.lt.0) rseed = rseed + im
    j = 1 + iy / ndiv
    iy = iv(j)
    iv(j) = rseed
    myran1 = min(am * iy, rnmx)

  end function myran1

  !---------------------------------------------------------------------
  subroutine rebinBasicReal(oldYvalues, newYvalues, status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) ::&
      oldYvalues(:)
    real, intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, newBin, localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    do oldBin = 1, numOldBins + 1
      oldXedges(oldBin) = (oldBin - 1) / real(numOldBins)
    end do
    do newBin = 1, numNewBins + 1
      newXedges(newBin) = (newBin - 1) / real(numNewBins)
    end do

    call rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
    , localStatus)

    if (present(status)) status = localStatus

  end subroutine rebinBasicReal

  !---------------------------------------------------------------------
  subroutine rebinBasicDble(oldYvalues, newYvalues, status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldYvalues(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, newBin, localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    do oldBin = 1, numOldBins + 1
      oldXedges(oldBin) = (oldBin - 1) / real(numOldBins)
    end do
    do newBin = 1, numNewBins + 1
      newXedges(newBin) = (newBin - 1) / real(numNewBins)
    end do

    call rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
    , localStatus)

    if (present(status)) status = localStatus

  end subroutine rebinBasicDble

  !---------------------------------------------------------------------
  subroutine rebinWithWidthsReal(oldYvalues, newYvalues, oldXwidths&
  , newXwidths, oldXcentres, newXcentres, status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in), dimension(:) ::&
      oldXwidths, oldXcentres, oldYvalues, newXwidths, newXcentres
    real, intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins,&
      localStatus, foreignStatus
    integer ::&
      oldBin, newBin

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXwidths) .lt. numOldBins) then
      if(present(status)) status = -1
  return
    else if (size(oldXwidths) .gt. numOldBins) then
      localStatus = localStatus + 1
    end if
    if (size(newXwidths) .lt. numNewBins) then
      if(present(status)) status = -2
  return
    else if (size(newXwidths) .gt. numNewBins) then
      localStatus = localStatus + 2
    end if

    if (size(oldXcentres) .lt. numOldBins) then
      if(present(status)) status = -3
  return
    else if (size(oldXcentres) .gt. numOldBins) then
      localStatus = localStatus + 4
    end if
    if (size(newXcentres) .lt. numNewBins) then
      if(present(status)) status = -4
  return
    else if (size(newXcentres) .gt. numNewBins) then
      localStatus = localStatus + 8
    end if

    do oldBin = 1, numOldBins
      oldXedges(oldBin) = oldXcentres(oldBin) - 0.5 * oldXwidths(oldBin)
    end do
    oldXedges(numOldBins + 1) = oldXcentres(numOldBins) + 0.5&
    * oldXwidths(numOldBins)

    do newBin = 1, numNewBins
      newXedges(newBin) = newXcentres(newBin) - 0.5 * newXwidths(newBin)
    end do
    newXedges(numNewBins + 1) = newXcentres(numNewBins) + 0.5&
    * newXwidths(numNewBins)

    call rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
    , foreignStatus)
  if (foreignStatus .lt. 0) return

    localStatus = localStatus + foreignStatus
    if(present(status)) status = localStatus

  end subroutine rebinWithWidthsReal

  !---------------------------------------------------------------------
  subroutine rebinWithWidthsDble(oldYvalues, newYvalues, oldXwidths&
  , newXwidths, oldXcentres, newXcentres, status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in), dimension(:) ::&
      oldXwidths, oldXcentres, oldYvalues, newXwidths, newXcentres
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins,&
      localStatus, foreignStatus
    integer ::&
      oldBin, newBin

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXwidths) .lt. numOldBins) then
      if(present(status)) status = -1
  return
    else if (size(oldXwidths) .gt. numOldBins) then
      localStatus = localStatus + 1
    end if
    if (size(newXwidths) .lt. numNewBins) then
      if(present(status)) status = -2
  return
    else if (size(newXwidths) .gt. numNewBins) then
      localStatus = localStatus + 2
    end if

    if (size(oldXcentres) .lt. numOldBins) then
      if(present(status)) status = -3
  return
    else if (size(oldXcentres) .gt. numOldBins) then
      localStatus = localStatus + 4
    end if
    if (size(newXcentres) .lt. numNewBins) then
      if(present(status)) status = -4
  return
    else if (size(newXcentres) .gt. numNewBins) then
      localStatus = localStatus + 8
    end if

    do oldBin = 1, numOldBins
      oldXedges(oldBin) = oldXcentres(oldBin) - 0.5d0 * oldXwidths(oldBin)
    end do
    oldXedges(numOldBins + 1) = oldXcentres(numOldBins) + 0.5d0&
    * oldXwidths(numOldBins)

    do newBin = 1, numNewBins
      newXedges(newBin) = newXcentres(newBin) - 0.5d0 * newXwidths(newBin)
    end do
    newXedges(numNewBins + 1) = newXcentres(numNewBins) + 0.5d0&
    * newXwidths(numNewBins)

    call rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
    , foreignStatus)
  if (foreignStatus .lt. 0) return

    localStatus = localStatus + foreignStatus
    if(present(status)) status = localStatus

  end subroutine rebinWithWidthsDble

  !---------------------------------------------------------------------
  subroutine rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------
    implicit none 

    real, intent(in) ::&
      oldXedges(:), oldYvalues(:), newXedges(:)
    real, intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    integer :: foreignStatus
    real(kind(1d0)) :: newYvaluesDble(size(newYvalues))

    !-------------------------------------------------------------------
    ! Start:

    call rebinKernel(dble(oldYvalues), newYvaluesDble, dble(oldXedges)&
    , dble(newXedges), foreignStatus)

    if (present(status)) status = foreignStatus
    newYvalues = real(newYvaluesDble)

  end subroutine rebinWithEdgesReal

  !---------------------------------------------------------------------
  subroutine rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------
    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldXedges(:), oldYvalues(:), newXedges(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start:

    call rebinKernel(oldYvalues, newYvalues, oldXedges&
    , newXedges, foreignStatus)

    if (present(status)) status = foreignStatus

  end subroutine rebinWithEdgesDble

  !---------------------------------------------------------------------
  subroutine rebinKernel(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------
  ! This is designed to be a pretty general and robust rebinning routine.
  ! Neither the old or new series is constrained to have bins of equal width.

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldXedges(:), oldYvalues(:), newXedges(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out) :: status

    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, oldStartBin, oldFinisBin, newBin
    logical ::&
      foundOverlappingBin
    real(kind(1d0)) ::&
      overlap, binWidth

    !-------------------------------------------------------------------
    ! Start:

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXedges) .lt. numOldBins + 1) then
      status = -1
  return
    else if (size(oldXedges) .gt. numOldBins + 1) then
      status = status + 1
    end if
    if (size(newXedges) .lt. numNewBins + 1) then
      status = -2
  return
    else if (size(newXedges) .gt. numNewBins + 1) then
      status = status + 2
    end if

    do oldBin = 1, numOldBins
      binWidth = oldXedges(oldBin + 1) - oldXedges(oldBin)
      if (binWidth .le. 0.0) then
        status = -3
  return
      end if
    end do

    do newBin = 1, numNewBins
      binWidth = newXedges(newBin + 1) - newXedges(newBin)
      if (binWidth .le. 0.0) then
        status = -4
  return
      end if
    end do

    newYvalues = 0.0 ! vector op, default.

    oldStartBin = 1
    NEW: do newBin = 1, numNewBins
      ! Find first old bin that overlaps newBin:
      foundOverlappingBin = .false. ! default
      OLD1: do oldBin = oldStartBin, numOldBins
        if (oldXedges(oldBin + 1) .gt. newXedges(newBin)&
        .and. oldXedges(oldBin) .lt. newXedges(newBin + 1)) then
          foundOverlappingBin = .true.
          oldStartBin = oldBin
      exit OLD1
        end if
      end do OLD1

      if (.not. foundOverlappingBin) then ! have either reached the end of the
      ! old series, or not reached the start. Leave new values at default 0.0.
    exit NEW
      end if

      ! Find the last old bin that overlaps newBin:
      foundOverlappingBin = .true. ! default
      oldFinisBin = numOldBins ! default
      OLD2: do oldBin = oldStartBin + 1, numOldBins
        if (oldXedges(oldBin) .ge. newXedges(newBin + 1)) then
          foundOverlappingBin = .false.
          oldFinisBin = oldBin - 1
      exit OLD2
        end if
      end do OLD2

      if (foundOverlappingBin) then ! have reached the end of the
      ! old series without leaving the new bin.
        oldFinisBin = numOldBins
      end if

      ! Now work though all the old bins that overlap the new, and pour in
      ! the correct proportional amounts from them into newBin:
      OLD3: do oldBin = oldStartBin, oldFinisBin
        if (oldXedges(oldBin + 1) .lt. newXedges(newBin + 1)) then
          if (oldXedges(oldBin) .gt. newXedges(newBin)) then !B - all
          ! of old bin goes into new:
            overlap = oldXedges(oldBin + 1) - oldXedges(oldBin)

          else !A - upper part of old goes into new:
            overlap = oldXedges(oldBin + 1) - newXedges(newBin)
          end if

        else
          if (oldXedges(oldBin) .gt. newXedges(newBin)) then !D - lower
          ! part of old goes into new:
            overlap = newXedges(newBin + 1) - oldXedges(oldBin)

          else !C - middle part of old goes into new:
            overlap = newXedges(newBin + 1) - newXedges(newBin)
          end if
        end if
        newYvalues(newBin) = newYvalues(newBin) + oldYvalues(oldBin)&
        * overlap
      end do OLD3

      oldStartBin = oldFinisBin
    end do NEW

    ! Divide by the new bin width:
    do newBin = 1, numNewBins
      binWidth = newXedges(newBin + 1) - newXedges(newBin)
      newYvalues(newBin) = newYvalues(newBin) / binWidth
    end do

  end subroutine rebinKernel

  !---------------------------------------------------------------------
  function normaliseReal(argument)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: argument(:)
    real :: normaliseReal(size(argument)), vectorLength2
    integer :: i

    !-------------------------------------------------------------------
    ! Start

    vectorLength2 = 0.0
    do i = 1, size(argument)
      vectorLength2 = vectorLength2 + argument(i) * argument(i)
    end do
    if (vectorLength2 .le. 0.0) then
      normaliseReal = 0.0
    else
      normaliseReal = argument / sqrt(vectorLength2)
    end if

  end function normaliseReal

  !---------------------------------------------------------------------
  function normaliseDble(argument)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), intent(in) :: argument(:)
    real(kind(1d0)) :: normaliseDble(size(argument)), vectorLength2
    integer :: i

    !-------------------------------------------------------------------
    ! Start

    vectorLength2 = 0.0d0
    do i = 1, size(argument)
      vectorLength2 = vectorLength2 + argument(i) * argument(i)
    end do
    if (vectorLength2 .le. 0.0) then
      normaliseDble = 0.0d0
    else
      normaliseDble = argument / sqrt(vectorLength2)
    end if

  end function normaliseDble

  !---------------------------------------------------------------------
  function vectorCrossReal(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real, dimension(3), intent(in) :: vectorA, vectorB
    real :: vectorCrossReal(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossReal(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossReal(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossReal(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossReal

  !---------------------------------------------------------------------
  function vectorCrossDble(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), dimension(3), intent(in) :: vectorA, vectorB
    real(kind(1d0)) :: vectorCrossDble(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossDble(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossDble(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossDble(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossDble

  !---------------------------------------------------------------------
  function bias(sag, x) result(y)
  !---------------------------------------------------------------------
  ! The action of this little function is to bias the x values (which should
  ! be between 0 and 1) either closer to 0 or closer to 1, depending on
  ! the value of 'sag'. For sag = -1, the 'upward' bias is at its limit:
  ! whatever the value of x, y will be set to 1. For sag = +1 the bias is
  ! at the opposite extreme, in which the function returns 0 for any value
  ! of x. For sag = 0, y = x. The interesting things happen when sag !=
  ! either -1, 0 or +1. In these cases the relation of y to x is a
  ! hyperbolic function that generates the required bias.
  !
  ! This function gives accurate results except for x values close to 1.
  ! A call such as
  !
  !	y - 1.0 - bias(-sag, 1.0-x)
  !
  ! is better in this instance.

    implicit none

    real(kind=single), intent(in) :: sag, x
    real(kind=single) :: y

    !-------------------------------------------------------------------
    ! Start:

    if (x .gt. 1.0 .or. x .lt. 0.0) then
      write(errstr, '(a, f10.7, a)') 'The argument ', x, ' is outside &
      &the range 0 to 1.'
call error('biasArgOutOfRange') ! should return a status really, and let the calling routine work out what to do.
    end if

    if (abs(sag) .gt. 1.0) then
      write(errstr, '(a, f10.7, a)') 'The sag value ', sag, ' is outside &
      &the range -1 to 1.'
call error('biasSagOutOfRange') ! should return a status really, and let the calling routine work out what to do.
    end if

    if (sag .eq. -1.0) then
      y = 0.0
    else if (sag .eq. 1.0) then
      y = 1.0
    else
      y = x * (1.0 + sag) / (1.0 + sag * (2.0 * x - 1.0))
    end if

  end function bias

  !---------------------------------------------------------------------
  function spread(stretch, x) result(y)
  !---------------------------------------------------------------------
  ! The action of this little function is to stretch or compress the x
  ! values (which should be between 0 and 1) about 0, depending on the
  ! value of 'stretch'. For stretch = -1, all the points are compressed
  ! into one: whatever the value of x, y will be set to 0. For stretch
  ! = +1 the spread is at its maximum, in which the function returns -1
  ! for negative values of x and +1 for positive values. At stretch = 0,
  ! y = x. The interesting things happen when stretch != either -1, 0
  ! or +1. In these cases the relation of y to x is a doUble hyperbolic
  ! function (generated by the appropriate call to the function bias)
  ! that generates the required spread.

    implicit none

    real(kind=single), intent(in) :: stretch, x
    real(kind=single) :: y

    !-------------------------------------------------------------------
    ! Start:

    if (abs(x) .gt. 1.0) then
      write(errstr, '(a, f10.7, a)') 'The argument ', x, ' is outside &
      &the range 0 to 1.'
call error('spreadArgOutOfRange') ! should return a status really, and let the calling routine work out what to do.
    end if

    if (abs(stretch) .gt. 1.0) then
      write(errstr, '(a, f10.7, a)') 'The stretch value ', stretch, ' is outside &
      &the range -1 to 1.'
call error('stretchOutOfRange') ! should return a status really, and let the calling routine work out what to do.
    end if

    if (stretch .eq. -1.0) then
      y = 0.0
    else if (stretch .eq. 1.0) then
      y = real(sign(1.0, x))
    else
      if (x .gt. 0.0) then
        y = bias(-stretch, x)
      else
        y = -bias(-stretch, -x)
      end if
    end if

  end function spread

  !---------------------------------------------------------------------
  function angstroms2eV(angstroms) result(EeV)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), intent(in) :: angstroms
    real(kind(1d0)) :: EeV

    !-------------------------------------------------------------------
    ! Start:

    EeV = PlanckH * SpeedOfLight / (angstroms * 1.0d-10) / ElectronQ

  end function angstroms2eV

  !---------------------------------------------------------------------
  function eV2angstroms(EeV) result(angstroms)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), intent(in) :: EeV
    real(kind(1d0)) :: angstroms

    !-------------------------------------------------------------------
    ! Start:

    angstroms = (PlanckH * SpeedOfLight / EeV / ElectronQ) * 1.0d10

  end function eV2angstroms

  !---------------------------------------------------------------------
  function bbFluxPerHertz_old(photonEnergyKeV, tempMegaKelvin) result(flux)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: photonEnergyKeV, tempMegaKelvin
    real :: flux, arg

    !-------------------------------------------------------------------
    ! Start:

!    arg = photonEnergyKeV * qeOnBoltzmann / tempMegaKelvin / 1000.0
    arg = photonEnergyKeV * ElectronQ / Boltzmann / tempMegaKelvin / 1000.0
    if (arg .gt. 20.0) then
      flux = 0.0
    else if (arg .lt. 0.01) then
      flux = 2.0 * arg * arg / (2.0 + arg) ! long-wavelength approx. using
      ! the first three terms of the exp function.
    else
      flux = arg * arg * arg / (exp(arg) - 1.0)
    end if

    ! Scaling is ignored because it will
    ! be normalised anyway by the calling routine.

  end function bbFluxPerHertz_old

  !---------------------------------------------------------------------
  function bbFluxPerHertz(hertz, kelvin) result(power)
  !---------------------------------------------------------------------
  ! This function calculates the power (in watts) radiated per unit
  ! frequency (in hertz) per unit surface area (m^2) per unit solid
  ! angle (sr) at frequency='hertz' by a black body at temperature='kelvin'.
  ! It is designed to be portable and as fast yet accurate as possible.
  ! The precision at low frequency is limited by the calculation of
  ! exp(k1 * f) - 1: for k1 * f less than minArg, a series expansion of
  ! (exp(arg) - 1) / arg is used. The minimum precision occurs at
  ! arg = minArg and is approximately equal to precision(1d0) + log10(arg).
  ! Note that minArg should NOT be greater than 1, otherwise the series
  ! will not converge.

    implicit none

    real(kind(1d0)), intent(in) :: hertz, kelvin
    real(kind(1d0)) :: power, arg, k1, k2, maxarg, minarg, minTermValue,&
      summ, term
    logical :: firstCall
    integer :: n

    save firstCall, k1, k2, maxArg, minArg, minTermValue
    data firstCall /.true./
    data minArg /0.1d0/

    !-------------------------------------------------------------------
    ! Start:

    if (firstCall) then
      k1 = PlanckH / Boltzmann / kelvin
      k2 = 2.0d0 * PlanckH / SpeedOfLight / SpeedOfLight
      maxArg = log(huge(1d0))
      minTermValue = 10**(-precision(1d0))

      firstCall = .false.
    end if

    arg = k1 * hertz
    if (arg .ge. maxArg) then
      power = 0.0d0

    else if (arg .le. minArg) then
      summ = 0.0d0
      term = 1.0d0
      n = 1
      do while(term .gt. minTermValue)
        summ = summ + term
        n = n + 1
        term = term * arg / real(n, kind(1d0))
      end do
      power = k2 * hertz * hertz * hertz / arg / summ

    else
      power = k2 * hertz * hertz * hertz / (exp(arg) - 1.0d0)
    end if

  end function bbFluxPerHertz

  !---------------------------------------------------------------------
  function bbFluxPerMetre(angstroms, kelvin) result(power)
  !---------------------------------------------------------------------
  ! This function calculates the power (in watts) radiated per unit
  ! wavelength (in metres) per unit surface area (m^2) per unit solid
  ! angle (sr) at wavelength='angstroms' by a black body at temperature =
  ! 'kelvin'. It is designed to be portable and as fast yet accurate as
  ! possible. The precision at long wavelengths is limited by the
  ! calculation of exp(k1/lambda) - 1: for k1/lambda less than minArg, a
  ! series expansion of (exp(arg) - 1) / arg is used. The minimum precision
  ! occurs at arg = minArg and is approximately equal to precision(1d0) +
  ! log10(arg). Note that minArg should NOT be greater than 1, otherwise
  ! the series will not converge.

    implicit none

    real(kind(1d0)), intent(in) :: angstroms, kelvin
    real(kind(1d0)) :: power, arg, k1, k2, maxarg, minarg, minTermValue,&
      summ, term
    logical :: firstCall
    integer :: n

    save firstCall, k1, k2, maxArg, minArg, minTermValue
    data firstCall /.true./
    data minArg /0.1d0/

    !-------------------------------------------------------------------
    ! Start:

    if (firstCall) then
      k1 = PlanckH * SpeedOfLight / 1.0d-10 / Boltzmann / kelvin
      k2 = 2.0d0 * PlanckH * SpeedOfLight * SpeedOfLight
      maxArg = log(huge(1d0))
      minTermValue = 10**(-precision(1d0))

      firstCall = .false.
    end if

    arg = k1 / angstroms
    if (arg .ge. maxArg) then
      power = 0.0d0

    else if (arg .le. minArg) then
      summ = 0.0d0
      term = 1.0d0
      n = 1
      do while(term .gt. minTermValue)
        summ = summ + term
        n = n + 1
        term = term * arg / real(n, kind(1d0))
      end do
      power = k2 / angstroms / angstroms / angstroms / angstroms / angstroms&
      / arg / summ

    else
      power = k2 / angstroms / angstroms / angstroms / angstroms / angstroms&
      / (exp(arg) - 1.0d0)
    end if

  end function bbFluxPerMetre

  !---------------------------------------------------------------------
  subroutine cmplxArrayDivideDble(srcImageA, srcImageB, xsize, ysize)
  !---------------------------------------------------------------------

! Divides A by B.

    implicit none 

    integer, intent(in) :: xsize, ysize
!    complex(kind=real(kind(1d0))), dimension(xsize, ysize), intent(inout) ::&
    complex(kind(1d0)), dimension(xsize, ysize), intent(inout) ::&
      srcImageA
!    complex(kind=real(kind(1d0))), dimension(xsize, ysize), intent(in) ::&
    complex(kind(1d0)), dimension(xsize, ysize), intent(in) ::&
      srcImageB
    real(kind(1d0)) :: denom, rPartA, iPartA, rPartB, iPartB, rPartC, iPartC
    integer :: xi, yi

!------------------------------------------------------------------------
! Start:

    do xi = 1, xsize
      do yi = 1, ysize
        rPartA = real(srcImageA(xi, yi))
        iPartA = aimag(srcImageA(xi, yi))
        rPartB = real(srcImageB(xi, yi))
        iPartB = aimag(srcImageB(xi, yi))
        denom = rPartB * rPartB + iPartB * iPartB
        if (denom .eq. 0.0) then
          srcImageA(xi, yi) = (0d0, 0d0)
        else
          rPartC = rPartA * rPartB + iPartA * iPartB
          iPartC = iPartA * rPartB - rPartA * iPartB
!          srcImageA(xi, yi) = cmplx(rPartC, iPartC, real(kind(1d0)))
          srcImageA(xi, yi) = cmplx(rPartC, iPartC, kind(1d0))
        end if
      end do
    end do

  end subroutine cmplxArrayDivideDble

  !---------------------------------------------------------------------
  subroutine cmplxArrayDivideReal(srcImageA, srcImageB, xsize, ysize)
  !---------------------------------------------------------------------

! Divides A by B.

    implicit none 

    integer, intent(in) :: xsize, ysize
    complex, dimension(xsize, ysize), intent(inout) ::&
      srcImageA
    complex, dimension(xsize, ysize), intent(in) ::&
      srcImageB
    real :: denom, rPartA, iPartA, rPartB, iPartB, rPartC, iPartC
    integer :: xi, yi

!------------------------------------------------------------------------
! Start:

    do xi = 1, xsize
      do yi = 1, ysize
        rPartA = real(srcImageA(xi, yi))
        iPartA = aimag(srcImageA(xi, yi))
        rPartB = real(srcImageB(xi, yi))
        iPartB = aimag(srcImageB(xi, yi))
        denom = rPartB * rPartB + iPartB * iPartB
        if (denom .eq. 0.0) then
          srcImageA(xi, yi) = (0.0, 0.0)
        else
          rPartC = rPartA * rPartB + iPartA * iPartB
          iPartC = iPartA * rPartB - rPartA * iPartB
          srcImageA(xi, yi) = cmplx(rPartC, iPartC)
        end if
      end do
    end do

  end subroutine cmplxArrayDivideReal

  !---------------------------------------------------------------------
  subroutine cmplxVectorDivideReal(vectorNum, vectorDen, vectorRes, size)
  !---------------------------------------------------------------------

! Divides vectorNum by vectorDen and leaves result in vectorRes.

    implicit none 

    integer, intent(in) :: size
    complex, dimension(size), intent(in) ::&
      vectorNum, vectorDen
    complex, dimension(size), intent(out) ::&
      vectorRes
    real :: denom, rPartA, iPartA, rPartB, iPartB, rPartC, iPartC
    integer :: i

!------------------------------------------------------------------------
! Start:

    do i = 1, size
      rPartA = real(vectorNum(i))
      iPartA = aimag(vectorNum(i))
      rPartB = real(vectorDen(i))
      iPartB = aimag(vectorDen(i))
      denom = rPartB * rPartB + iPartB * iPartB
      if (denom .eq. 0.0) then
        vectorRes(i) = (0.0, 0.0)
      else
        rPartC = rPartA * rPartB + iPartA * iPartB
        iPartC = iPartA * rPartB - rPartA * iPartB
        vectorRes(i) = cmplx(rPartC, iPartC)
      end if
    end do

  end subroutine cmplxVectorDivideReal

  !---------------------------------------------------------------------
  subroutine integrate1dRealIntervals(y, intervals, n, result, status)
  !---------------------------------------------------------------------

    implicit none

    type(intervalsType)            :: intervals(:)
    integer, intent(in),  optional :: n
    real,    intent(out)           :: result
    integer, intent(out), optional :: status

    real :: realNumMidSamples, frac,&
      x, midIntegral, localLo, localHi
    real, allocatable :: inInterval(:)
    integer :: i, localStatus, localN, numMidSamples, loN, hiN

    interface
      real function y(x)
        real, intent(in) :: x
      end function y
    end interface

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    if (present(n)) then
      localN = n
    else
      localN = defaultNumIntegralSamples
    end if

    !**** should check intervals are in order, and don't overlap.

    !**** can't yet handle infinite bounds.

localLo = minval(intervals%lower%value)
localHi = maxval(intervals%upper%value)
allocate(inInterval(localN))
inInterval = 0.0 ! default
do i = 1, size(intervals)
  loN = 1 + nint(localN * ((intervals(i)%lower%value - localLo) / (localHi - localLo)))
  hiN = 1 + nint(localN * ((intervals(i)%upper%value - localLo) / (localHi - localLo)))
  loN = min(loN, localN)
  hiN = min(hiN, localN)
  inInterval(loN:hiN) = 1.0
end do
numMidSamples = localN
    realNumMidSamples = real(numMidSamples)

    midIntegral = 0.0
    do i = 1, numMidSamples
      frac = (i - 0.5) / realNumMidSamples
      x = localLo * (1.0 - frac) + localHi * frac
midIntegral = midIntegral + y(x) * inInterval(i)
    end do
    midIntegral = midIntegral * (localHi - localLo) / realNumMidSamples

    result = midIntegral

    if (present(status)) status = localStatus

  end subroutine integrate1dRealIntervals

  !---------------------------------------------------------------------
  subroutine integrate1dReal(y, lo, hi, loInf, hiInf, n, result, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in),  optional :: lo, hi
    logical, intent(in),  optional :: loInf, hiInf
    integer, intent(in),  optional :: n
    real,    intent(out)           :: result
    integer, intent(out), optional :: status

    real :: realNumLoSamples, realNumMidSamples, realNumHiSamples, frac,&
      x, loIntegral, midIntegral, hiIntegral, localLo, localHi, u
    integer :: i, localStatus, localN, numLoSamples, numMidSamples, numHiSamples
    logical :: localLoInf, localHiInf

    interface
      real function y(x)
        real, intent(in) :: x
      end function y
    end interface

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    if (present(lo)) then
      localLo = lo
      if (present(loInf)) then
        localLoInf = loInf
      else
        localLoInf = .false.
      end if
      if (localLoInf .and. localLo .gt. 0.0) then
        if (present(status)) status = -1
  return
      end if
    else
      localLoInf = .true.
      localLo = -1.0
    end if

    if (present(hi)) then
      localHi = hi
      if (present(hiInf)) then
        localHiInf = hiInf
      else
        localHiInf = .false.
      end if
      if (localHiInf .and. localHi .lt. 0.0) then
        if (present(status)) status = -2
  return
      end if
    else
      localHiInf = .true.
      localHi = 1.0
    end if

    if (present(n)) then
      localN = n
    else
      localN = defaultNumIntegralSamples
    end if

    if (present(lo) .and. present(hi)) then
      if (lo .gt. hi) then
        if (present(status)) status = -3
  return
      else if (lo .eq. hi) then
        result = 0.0
        localStatus = localStatus + 1
        if (present(status)) status = localStatus
  return
      end if
    end if

    if (localLoInf) then
      if (localHiInf) then
        numLoSamples = floor(localN / 3.0)
        numHiSamples = floor(localN / 3.0)
        if (numLoSamples .lt. 1) then
          if (present(status)) status = -4
  return
        end if
      else
        numHiSamples = 0
        numLoSamples = floor(localN / 2.0)
        if (numLoSamples .lt. 1) then
          if (present(status)) status = -5
  return
        end if
      end if
    else
      numLoSamples = 0
      if (localHiInf) then
        numHiSamples = floor(localN / 2.0)
        if (numHiSamples .lt. 1) then
          if (present(status)) status = -6
  return
        end if
      else
        numHiSamples = 0
      end if
    end if
    numMidSamples = localN - numLoSamples - numHiSamples
    if (numMidSamples .lt. 1) then
      if (present(status)) status = -7
  return
    end if

    realNumLoSamples = real(numLoSamples)
    realNumMidSamples = real(numMidSamples)
    realNumHiSamples = real(numHiSamples)

    loIntegral = 0.0
    if (realNumLoSamples .gt. 0) then
      do i = 1, numLoSamples
        frac = (i - 0.5) / realNumLoSamples
        u = frac / localLo
        x = 1.0 / u
        loIntegral = loIntegral + y(x) / u / u
      end do
      loIntegral = loIntegral / realNumLoSamples / abs(localLo)
    end if

    midIntegral = 0.0
    do i = 1, numMidSamples
      frac = (i - 0.5) / realNumMidSamples
      x = localLo * (1.0 - frac) + localHi * frac
      midIntegral = midIntegral + y(x)
    end do
    midIntegral = midIntegral * (localHi - localLo) / realNumMidSamples

    hiIntegral = 0.0
    if (realNumHiSamples .gt. 0) then
      do i = 1, numHiSamples
        frac = (i - 0.5) / realNumHiSamples
        u = frac / localHi
        x = 1.0 / u
        hiIntegral = hiIntegral + y(x) / u / u
      end do
      hiIntegral = hiIntegral / realNumHiSamples / localHi
    end if

    result = loIntegral + midIntegral +  hiIntegral

    if (present(status)) status = localStatus

  end subroutine integrate1dReal

  !---------------------------------------------------------------------
  subroutine integrateNdReal(y, args, freeIndices, lo, hi, loInf, hiInf, n&
  , dims, result, status)
  !---------------------------------------------------------------------
! traps: actual y called with x(some number) instead of x(:); setting all lo/hiinf() but not all lo/hi().

    implicit none

    real,    intent(in),  optional :: lo(:), hi(:), args(:)
    logical, intent(in),  optional :: loInf(:), hiInf(:), freeIndices(:)
    integer, intent(in),  optional :: n(:)
    integer, intent(in)            :: dims
    real,    intent(out)           :: result
    integer, intent(out), optional :: status

    real :: localLo(dims), localHi(dims), localArgs(dims)
    integer :: localStatus, numLoSamples(dims), numMidSamples(dims),&
      numHiSamples(dims), localN(dims), i
    logical :: localLoInf(dims), localHiInf(dims), localFreeIndices(dims),&
      infClashFlag

    interface
      real function y(x)
        real, intent(in) :: x(:)
      end function y
    end interface

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

!***** check all sizes are equal to dims

    if (present(lo)) then
      localLo = lo
      if (present(loInf)) then
        localLoInf = loInf
      else
        localLoInf = .false.
      end if
      do i = 1, dims
        if (localLoInf(i) .and. localLo(i) .ge. 0.0) then
          if (present(status)) status = -1
  return
        end if
      end do
    else
      localLo = -1.0
      localLoInf = .true.
      if (present(loInf)) then
        infClashFlag = .false. ! default
        do i = 1, dims
          if (.not. loInf(i)) then
            infClashFlag = .true.
        exit
          end if
        end do
        if (infClashFlag .and. present(status)) status = status + 2
      end if
    end if

    if (present(hi)) then
      localHi = hi
      if (present(hiInf)) then
        localHiInf = hiInf
      else
        localHiInf = .false.
      end if
      do i = 1, dims
        if (localHiInf(i) .and. localHi(i) .le. 0.0) then
          if (present(status)) status = -2
  return
        end if
      end do
    else
      localHi = 1.0
      localHiInf = .true.
      if (present(hiInf)) then
        infClashFlag = .false. ! default
        do i = 1, dims
          if (.not. hiInf(i)) then
            infClashFlag = .true.
        exit
          end if
        end do
        if (infClashFlag .and. present(status)) status = status + 4
      end if
    end if

    if (present(args)) then
      localArgs = args
      if (present(freeIndices)) then
        localFreeIndices = freeIndices
      else
        localFreeIndices = .true.
      end if
    else
      localArgs = 0.0
      localFreeIndices = .true.
    end if

    if (present(n)) then
      localN = n
    else
      localN = defaultNumIntegralSamples
    end if

    if (present(lo) .and. present(hi)) then
      do i = 1, dims
        if (lo(i) .gt. hi(i)) then
        if (present(status)) status = -3
  return
        else if (lo(i) .eq. hi(i)) then
          result = 0.0
          localStatus = localStatus + 1
          if (present(status)) status = localStatus
  return
        end if
      end do
    end if

    do i = 1, dims
      if (localLoInf(i)) then
        if (localHiInf(i)) then
          numLoSamples(i) = floor(localN(i) / 3.0)
          numHiSamples(i) = floor(localN(i) / 3.0)
          if (numLoSamples(i) .lt. 1) then
            if (present(status)) status = -4
  return
          end if
        else
          numHiSamples(i) = 0
          numLoSamples(i) = floor(localN(i) / 2.0)
          if (numLoSamples(i) .lt. 1) then
            if (present(status)) status = -5
  return
          end if
        end if
      else
        numLoSamples(i) = 0
        if (localHiInf(i)) then
          numHiSamples(i) = floor(localN(i) / 2.0)
          if (numHiSamples(i) .lt. 1) then
            if (present(status)) status = -6
  return
          end if
        else
          numHiSamples(i) = 0
        end if
      end if
      numMidSamples(i) = localN(i) - numLoSamples(i) - numHiSamples(i)
      if (numMidSamples(i) .lt. 1) then
        if (present(status)) status = -7
  return
      end if
    end do

    call integrateKernel(y, localArgs, localFreeIndices, localLo&
    , localHi, numLoSamples, numMidSamples, numHiSamples, result&
    , localStatus)

    if (present(status)) status = localStatus

  end subroutine integrateNdReal

  !---------------------------------------------------------------------
  recursive subroutine integrateKernel(y, args, freeIndices, lo, hi&
  , numLoSamples, numMidSamples, numHiSamples, result, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)  :: lo(:), hi(:), args(:)
    logical, intent(in)  :: freeIndices(:)
    integer, intent(in)  :: numLoSamples(:), numMidSamples(:), numHiSamples(:)
    real,    intent(out) :: result
    integer, intent(out) :: status

    real :: realNumLoSamples, realNumMidSamples, realNumHiSamples, frac,&
      x(size(args)), u, localLo, localHi, loIntegral, midIntegral, hiIntegral,&
      localResult, term
    integer :: i, dims, firstFreeIndex, lastFreeIndex
    logical :: localFreeIndices(size(args))

    interface
      real function y(x)
        real, intent(in) :: x(:)
      end function y
    end interface

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default
    dims = size(args)

    ! Find index of last free index:
    do i = dims, 1, -1
      if (freeIndices(i)) then
        lastFreeIndex = i
    exit
      end if
    end do

    ! Find index of first free index and integrate over that:
    do i = 1, dims
      if (freeIndices(i)) then
        firstFreeIndex = i
    exit
      end if
    end do

    realNumLoSamples = real(numLoSamples(firstFreeIndex))
    realNumMidSamples = real(numMidSamples(firstFreeIndex))
    realNumHiSamples = real(numHiSamples(firstFreeIndex))
    localLo = lo(firstFreeIndex)
    localHi = hi(firstFreeIndex)

    x = args

    if (firstFreeIndex .eq. lastFreeIndex) then ! direct integration:
      loIntegral = 0.0
      if (realNumLoSamples .gt. 0.0) then
        do i = 1, numLoSamples(firstFreeIndex)
          frac = (i - 0.5) / realNumLoSamples
          u = frac / localLo
          x(firstFreeIndex) = 1.0 / u
          term = y(x) / u / u
          loIntegral = loIntegral + term
        end do
        loIntegral = loIntegral / realNumLoSamples / abs(localLo)
      end if

      midIntegral = 0.0
      do i = 1, numMidSamples(firstFreeIndex)
        frac = (i - 0.5) / realNumMidSamples
        x(firstFreeIndex) = localLo * (1.0 - frac) + localHi * frac
        term = y(x)
        midIntegral = midIntegral + term
      end do
      midIntegral = midIntegral * (localHi - localLo) / realNumMidSamples

      hiIntegral = 0.0
      if (realNumHiSamples .gt. 0) then
        do i = 1, numHiSamples(firstFreeIndex)
          frac = (i - 0.5) / realNumHiSamples
          u = frac / localHi
          x(firstFreeIndex) = 1.0 / u
          term = y(x) / u / u
          hiIntegral = hiIntegral + term
        end do
        hiIntegral = hiIntegral / realNumHiSamples / localHi
      end if

    else ! recursive call:
      localFreeIndices = freeIndices
      localFreeIndices(firstFreeIndex) = .false.

      loIntegral = 0.0
      if (realNumLoSamples .gt. 0.0) then
        do i = 1, numLoSamples(firstFreeIndex)
          frac = (i - 0.5) / realNumLoSamples
          u = frac / localLo
          x(firstFreeIndex) = 1.0 / u
          call integrateKernel(y, x, localFreeIndices, lo&
          , hi, numLoSamples, numMidSamples, numHiSamples, localResult&
          , status)

          loIntegral = loIntegral + localResult / u / u
        end do
        loIntegral = loIntegral / realNumLoSamples / abs(localLo)
      end if

      midIntegral = 0.0
      do i = 1, numMidSamples(firstFreeIndex)
        frac = (i - 0.5) / realNumMidSamples
        x(firstFreeIndex) = localLo * (1.0 - frac) + localHi * frac
        call integrateKernel(y, x, localFreeIndices, lo&
        , hi, numLoSamples, numMidSamples, numHiSamples, localResult&
        , status)

        midIntegral = midIntegral + localResult
      end do
      midIntegral = midIntegral * (localHi - localLo) / realNumMidSamples

      hiIntegral = 0.0
      if (realNumHiSamples .gt. 0) then
        do i = 1, numHiSamples(firstFreeIndex)
          frac = (i - 0.5) / realNumHiSamples
          u = frac / localHi
          x(firstFreeIndex) = 1.0 / u
          call integrateKernel(y, x, localFreeIndices, lo&
          , hi, numLoSamples, numMidSamples, numHiSamples, localResult&
          , status)

          hiIntegral = hiIntegral + localResult / u / u
        end do
        hiIntegral = hiIntegral / realNumHiSamples / localHi
      end if
    end if

    result = loIntegral + midIntegral +  hiIntegral

  end subroutine integrateKernel

  !---------------------------------------------------------------------
  subroutine newtonsMethod(startX, y, dy, solutionX, status)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: startX
    real, intent(out) :: solutionX
    integer, intent(out) :: status

    interface
      real function y(x)
        real, intent(in) :: x
      end function y
      real function dy(x)
        real, intent(in) :: x
      end function dy
    end interface

    real :: yValue, yDeriv, x, oldXstep, xStep, myPrecision

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    myPrecision = 10.0 ** (-precision(x))

    yValue = y(startX)
    yDeriv = dy(startX)
    if (yDeriv .eq. 0.0) then
      status = -1
  return
    end if
    Xstep = -Yvalue / yDeriv
    if (abs(startX) .gt. 0.0) then
      if (abs(Xstep / startX) .le. myPrecision) then
        solutionX = startX + Xstep
  return
      end if
    end if
    x = startX + Xstep
    oldXstep = Xstep

    do
      yValue = y(x)
      yDeriv = dy(x)
      if (yDeriv .eq. 0.0) then
        status = -1
  return
      end if

      Xstep = -yValue / yDeriv

      if (abs(x) .gt. 0.0) then
        if (abs(Xstep / x) .le. myPrecision) then
          solutionX = x + Xstep
  return
        end if
      end if

      if (abs(Xstep) .gt. abs(oldXstep)) then
        status = -2
  return
      end if

      x = x + Xstep
      oldXstep = Xstep
    end do

  end subroutine newtonsMethod

  !---------------------------------------------------------------------
  subroutine interpolate(xVector, yVector, x, y, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)            :: xVector(:), yVector(:), x
    real,    intent(out)           :: y
    integer, intent(out), optional :: status

    integer :: localStatus, numPoints, i, j
    logical :: xPointsAreInOrder

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    numPoints = size(xVector)

    if (size(yVector) .ne. numPoints) then
      localStatus = -1

    else
      xPointsAreInOrder = .true.
      do i = 1, numPoints - 1
        if (xVector(i) .ge. xVector(i + 1)) then
          xPointsAreInOrder = .false.
      exit
        end if
      end do
      if (.not. xPointsAreInOrder) then
        localStatus = -2

      else
        if (x .lt. xVector(1)) then
          localStatus = localStatus + 1
          y = 0.0

        else if (x .gt. xVector(numPoints)) then
          localStatus = localStatus + 2
          y = 0.0

        else if (numPoints .lt. 2) then
          localStatus = localStatus + 4
          y = yVector(1)

        else
          j = numPoints
          do i = 2, numPoints
            if (x .lt. xVector(i)) then
              j = i
          exit
            end if
          end do

          y = yVector(j - 1) + (x - xVector(j - 1))&
          * (yVector(j) - yVector(j - 1)) / (xVector(j) - xVector(j - 1))
        end if
      end if
    end if

    if (present(status)) status = localStatus

  end subroutine interpolate

  !---------------------------------------------------------------------
  subroutine getIntervalsANDandOR(intervals, ANDedIntervals, ORRedIntervals, status)
  !---------------------------------------------------------------------

    implicit none

    type(intervalsType), intent(in)            :: intervals(:)
    type(intervalType), dimension(:), pointer :: ANDedIntervals, ORRedIntervals
    integer,            intent(out), optional :: status

    type(intervalType) :: tempAI(size(intervals)), tempOI(size(intervals))
    integer :: i, j, minIndex, maxIndex, maxNumIndices, numIndices,&
      numEvents, numANDedIntervals, numORRedIntervals, numSimultaneousIntervals
    type(sequenceType) :: sequence(2*size(intervals))
    integer, allocatable :: numOverlaps(:), indicesFound(:)
    logical :: indexNotSeenAtPreviousI
!  type :: boundType
!    integer :: type
!    real(kind(1d0)) :: value
!  end type boundType
!  type :: intervalType
!    type(boundType) :: lower, upper
!    integer :: index
!  end type intervalType
!  type :: sequenceType
!    type(boundType) :: bound
!    logical         :: isLower
!    integer         :: index
!  end type sequenceType
    !-------------------------------------------------------------------
    ! Start:

!    emptyANDlist = .false. ! default
    status = 0 ! default
    if (size(intervals) .le. 0) then
      allocate(ANDedIntervals(1:0)) ! associate pointer with a zero-size array
      allocate(ORRedIntervals(1:0)) ! associate pointer with a zero-size array
  return
    end if

    ! Find out how many different index values there are:
    minIndex = minval(intervals%index)
    maxIndex = maxval(intervals%index)
    maxNumIndices = maxIndex - minIndex + 1
    allocate(numOverlaps(minIndex:maxIndex), indicesFound(maxNumIndices))
    numIndices = 0
    do i = 1, size(intervals)
      indexNotSeenAtPreviousI = .true.
      do j = 1, numIndices
        if (intervals(i)%index .eq. indicesFound(j)) then
          indexNotSeenAtPreviousI = .false.
      exit
        end if
      end do
      numIndices = numIndices + 1
      indicesFound(numIndices) = intervals(i)%index
    end do

    ! Load sequence:
    do i = 1, size(intervals)
      sequence(2*i-1)%bound%type = intervals(i)%lower%type
      sequence(2*i-1)%bound%value = intervals(i)%lower%value
      sequence(2*i-1)%isLower = .true.
      sequence(2*i-1)%index = intervals(i)%index

      sequence(2*i)%bound%type = intervals(i)%upper%type
      sequence(2*i)%bound%value = intervals(i)%upper%value
      sequence(2*i)%isLower = .false.
      sequence(2*i)%index = intervals(i)%index
    end do

    ! Sort sequence elements in order of increasing time (all isLower = .true. UNDEFINEDs at the start, all isLower = .false. UNDEFINEDs at the end):
    call sortSequence(sequence)

    ! Get rid of overlaps for each index:
    numOverlaps = 0 ! default
    numEvents = 0
    do i = 1, size(sequence)
      if (sequence(i)%isLower) then
        if (numOverlaps(sequence(i)%index) .eq. 0) then
          numEvents = numEvents + 1
          sequence(numEvents) = sequence(i)
        end if
        numOverlaps(sequence(i)%index) = numOverlaps(sequence(i)%index) + 1

      else ! sequence(i)%isUpper
        if (numOverlaps(sequence(i)%index) .eq. 1) then
          numEvents = numEvents + 1
          sequence(numEvents) = sequence(i)
        end if
        numOverlaps(sequence(i)%index) = numOverlaps(sequence(i)%index) - 1
      end if
    end do
!**** put in some bug detection for a while?

    ! Now the meat of the routine - calculation of the anded and orred sets of intervals:
!    tempAI%index = 0
!    tempOI%index = 0
    numANDedIntervals = 0
    numORRedIntervals  = 0
    numSimultaneousIntervals = 0
    do i = 1, numEvents
      if (sequence(i)%isLower) then
        numSimultaneousIntervals = numSimultaneousIntervals + 1
        if (numSimultaneousIntervals .gt. numIndices) then
!          if (present(status)) status = -200 ! actually a bug - shouldn't
!          ! happen.
!  return
        end if

        if (numSimultaneousIntervals .eq. 1) then
          numORRedIntervals = numORRedIntervals + 1
          tempOI(numORRedIntervals)%lower = sequence(i)%bound
!          tempOI(numORRedIntervals)%index = sequence(i)%index

        else if (numSimultaneousIntervals .eq. numIndices) then
          numANDedIntervals = numANDedIntervals + 1
          tempAI(numANDedIntervals)%lower = sequence(i)%bound
!          tempAI(numANDedIntervals)%index = sequence(i)%index
        end if

      else
        if (numSimultaneousIntervals .eq. 1) then
!          numORRedIntervals = numORRedIntervals + 1
          tempOI(numORRedIntervals)%upper = sequence(i)%bound
!          tempOI(numORRedIntervals)%index = sequence(i)%index

        else if (numSimultaneousIntervals .eq. numIndices) then
!          numANDedIntervals = numANDedIntervals + 1
          tempAI(numANDedIntervals)%upper = sequence(i)%bound
!          tempAI(numANDedIntervals)%index = sequence(i)%index
        end if

        numSimultaneousIntervals = numSimultaneousIntervals - 1
        if (numSimultaneousIntervals .lt. 0) then
!          if (present(status)) status = -300 ! actually a bug - shouldn't
!          ! happen.
!  return
        end if
      end if
    end do

    ! Get rid of abutting intervals:
    j = 1
    do i = 1, numANDedIntervals - 1
      if (.not. ((tempAI(j)%upper%type .eq. UNDEFINED&
      .or. tempAI(i + 1)%upper%type .eq. UNDEFINED)&
      .or. (tempAI(j)%upper%type .eq. EXCLUSIVE&
      .and. tempAI(i + 1)%upper%type .eq. EXCLUSIVE))&
      .and. tempAI(j)%upper%value .eq. tempAI(i + 1)%lower%value) then
      ! Delete the pair of equal times:
        tempAI(j)%upper = tempAI(i + 1)%upper
      else
        j = j + 1
        tempAI(j)%lower = tempAI(i + 1)%lower
        tempAI(j)%upper = tempAI(i + 1)%upper
      end if
    end do

!    if (present(ANDedIntervals)) then
      allocate(ANDedIntervals(j))
      ANDedIntervals = tempAI(1:j)
!    end if

    j = 1
    do i = 1, numORRedIntervals - 1
      if (.not. ((tempOI(j)%upper%type .eq. UNDEFINED&
      .or. tempOI(i + 1)%upper%type .eq. UNDEFINED)&
      .or. (tempOI(j)%upper%type .eq. EXCLUSIVE&
      .and. tempOI(i + 1)%upper%type .eq. EXCLUSIVE))&
      .and. tempOI(j)%upper%value .eq. tempOI(i + 1)%lower%value) then
      ! Delete the pair of equal times:
        tempOI(j)%upper = tempOI(i + 1)%upper
      else
        j = j + 1
        tempOI(j)%lower = tempOI(i + 1)%lower
        tempOI(j)%upper = tempOI(i + 1)%upper
      end if
    end do

!    if (present(ORRedIntervals)) then
      allocate(ORRedIntervals(j))
      ORRedIntervals = tempOI(1:j)
!    end if

  end subroutine getIntervalsANDandOR

  !---------------------------------------------------------------------
  subroutine sortSequence(sequence)
  !---------------------------------------------------------------------

    implicit none

    type(sequenceType), intent(inout) :: sequence(:)

    integer :: numDefinedBounds, numLowerUndefinedBounds, i, j, indices(size(sequence))
    type(sequenceType) :: tempSequence(size(sequence))
    real(kind(1d0)) :: events(size(sequence))

!  type :: boundType
!    integer :: type
!    real(kind(1d0)) :: value
!  end type boundType
!  type :: intervalType
!    type(boundType) :: lower, upper
!    integer :: index
!  end type intervalType
!  type :: sequenceType
!    type(boundType) :: bound
!    logical         :: isLower
!    integer         :: index
!  end type sequenceType
    !-------------------------------------------------------------------
    ! Start:

    numDefinedBounds = 0
    numLowerUndefinedBounds = 0
    do i = 1, size(sequence)
      if (sequence(i)%bound%type .ne. UNDEFINED) then
        numDefinedBounds = numDefinedBounds + 1
        indices(numDefinedBounds) = i
        events(numDefinedBounds) = sequence(i)%bound%value
      else if (sequence(i)%isLower) then
        numLowerUndefinedBounds = numLowerUndefinedBounds + 1
        tempSequence(numLowerUndefinedBounds) = sequence(i)
      end if
    end do

    if (numDefinedBounds .gt. 0) then
      call quick_sort(events(1:numDefinedBounds), indices(1:numDefinedBounds))
    end if

    do i = 1, numDefinedBounds
      tempSequence(numLowerUndefinedBounds + i) = sequence(indices(i))
    end do

    j = numLowerUndefinedBounds + numDefinedBounds
    do i = 1, size(sequence)
      if (sequence(i)%bound%type .eq. UNDEFINED&
      .and. .not.sequence(i)%isLower) then
        j = j + 1
        tempSequence(j) = sequence(i)
      end if
    end do

    sequence = tempSequence

  end subroutine sortSequence

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module mathutils_mod
